/* EXTENSION FILE */

/* nfadata - NFA named elements data routines */

#include "flexdef.h"

/**********************************************************
 * Methods for RegularExpressions (regExp)
 **********************************************************/

int newRegExp PROTO((int));
int createMinMaxOcc PROTO((int,int,int,int));

void regExp_closure_initialLevelManagement PROTO((int,int));
void regExp_closure_enableActionSet PROTO((int,int,NameLevel*,int));
int regExp_closure_setPriorities PROTO((int,int,int*,int));

void regExp_chain_enableActionSet PROTO((int,NameLevel*,int));
void regExp_union_enableActionSet PROTO((int,NameLevel*,int));
void regExp_named_enableActionSet PROTO((int,NameLevel*,int));

void regExp_enableActionSet PROTO((int,NameLevel*,int));
int regExp_setPriorities PROTO((int,int));

int machine_addActionTransitions PROTO((int,ActionSet*));

void regExp_top_levelManagement PROTO((int,int));

void finish_name_options PROTO(());
void finish_name_option PROTO((int*,char*,char*,int));

#define APPEND(name) (name_append[name] == NAME_OPTION_YES)
#define CONCAT(name) (name_concat[name] == NAME_OPTION_YES)

///////////////
// PUBLIC
///////////////

///////////////
// CONSTRUCTORS
///////////////

int newEpsilon() {
	return newRegExp(EPSILON);
}

int newValue(int value) {
	int index = newRegExp(VALUE);
	reg_exps[index].data.valueData.value = value;
	return index;
}

int newChain(int first_index, int second_index) {
	int index = newRegExp(CHAIN);
	RegExp *newOne = &reg_exps[index];
	RegExp *firstSubRegExp = &reg_exps[first_index];
	RegExp *secondSubRegExp = &reg_exps[second_index];
	
	NameSet *nameSet = &newOne->nameSet;
	NameSet *firstSubNameSet = &firstSubRegExp->nameSet;
	NameSet *secondSubNameSet = &secondSubRegExp->nameSet;
	
	newOne->data.chainData.firstSubRegExp = first_index;
	newOne->data.chainData.secondSubRegExp = second_index;
	nameSet_addAll(nameSet,firstSubNameSet);
	nameSet_addAll(nameSet,secondSubNameSet);
	return index;
}

int newUnion(int first_index,int second_index) {
	int i,n,index = newRegExp(UNION);
	RegExp *newOne = &reg_exps[index];
	newOne->data.unionData.firstSubRegExp = first_index;
	newOne->data.unionData.secondSubRegExp = second_index;
	
	RegExp *firstSubRegExp = &reg_exps[first_index];
	RegExp *secondSubRegExp = &reg_exps[second_index];
	
	NameSet *nameSet = &newOne->nameSet;
	NameSet *firstSubNameSet = &firstSubRegExp->nameSet;
	NameSet *secondSubNameSet = &secondSubRegExp->nameSet;
	
	int firstNameSetSize = firstSubNameSet->size;
	int secondNameSetSize = secondSubNameSet->size;
	
	int usedNamesCount = 0;
	int *usedNames = allocate_integer_array(firstNameSetSize);
	
	for(i=0; i<firstNameSetSize; i++) {
		NameLevelCount* nlc = &firstSubNameSet->array[i];
		int name = nlc->name;
		
		for(n=0; n<usedNamesCount; n++) {
			if(name == usedNames[n]) {
				name = -1;
				break;
			}
		}
		if(name == -1) continue;
		
		usedNames[usedNamesCount++] = name;
		
		NameLevelCount *secondSubMax = nameSet_getMaximumLevel(secondSubNameSet,name);
		
		if(secondSubMax != NIL) {
			NameLevelCount *firstSubMax = nameSet_getMaximumLevel(firstSubNameSet,name);
			
			int firstSubMaxLevel = firstSubMax->level;
			int secondSubMaxLevel = secondSubMax->level;
			
			int firstSubMaxCount = firstSubMax->count;
			int secondSubMaxCount = secondSubMax->count;
			
			if(firstSubMaxLevel > secondSubMaxLevel) {
				nameSet_addAll(nameSet,firstSubNameSet);
			}
			else if(secondSubMaxLevel > firstSubMaxLevel) {
				nameSet_addAll(nameSet,secondSubNameSet);
			}
			else {
				if(firstSubMaxCount > secondSubMaxCount) {
					nameSet_addNameLevelCount(nameSet,name,firstSubMaxLevel,firstSubMaxCount);
				}
				else if(secondSubMaxCount > firstSubMaxCount) {
					nameSet_addNameLevelCount(nameSet,name,secondSubMaxLevel,secondSubMaxCount);
				}
				else if(firstSubMaxCount > 1) {
					nameSet_addNameLevelCount(nameSet,name,firstSubMaxLevel,firstSubMaxCount);
				}
				else {
					NameLevelCount *firstSubSecondMax = nameSet_getSecondMaximumLevel(firstSubNameSet,name);
					NameLevelCount *secondSubSecondMax = nameSet_getSecondMaximumLevel(secondSubNameSet,name);
					
					if(firstSubSecondMax != NIL) {
						nameSet_addNameLevel(nameSet,name,firstSubMaxLevel);
						nameSet_addNameLevelCount(nameSet,name,firstSubSecondMax->level,firstSubSecondMax->count);
					}
					else if(secondSubSecondMax != NIL) {
						nameSet_addNameLevel(nameSet,name,secondSubMaxLevel);
						nameSet_addNameLevelCount(nameSet,name,secondSubSecondMax->level,secondSubSecondMax->count);
					}
					else {
						nameSet_addNameLevel(nameSet,name,firstSubMaxLevel);
					}
				}
			}
		}
		else nameSet_addAll_ofName(nameSet,firstSubNameSet,name);
	}
	
	for(i=0; i<secondNameSetSize; i++) {
		NameLevelCount* nlc = &secondSubNameSet->array[i];
		int name = nlc->name;
		
		for(n=0; n<usedNamesCount; n++) {
			if(name == usedNames[n]) {
				name = -1;
				break;
			}
		}
		if(name == -1) continue;
		
		nameSet_addAll_ofName(nameSet,secondSubNameSet,name);
	}
	
	flex_free(usedNames);
	
	return index;
}

int newStar(int subRegExp) {
	int index = newRegExp(STAR);
	RegExp *newOne = &reg_exps[index];
	newOne->data.starData.subRegExp = subRegExp;
	regExp_closure_initialLevelManagement(index,subRegExp);
	return index;
}

int newPlus(int subRegExp) {
	int index = newRegExp(PLUS);
	RegExp *newOne = &reg_exps[index];
	newOne->data.plusData.subRegExp = subRegExp;
	regExp_closure_initialLevelManagement(index,subRegExp);
	return index;
}

int newOptional(int sub_index) {
	int index = newRegExp(OPTIONAL);
	RegExp *newOne = &reg_exps[index];
	RegExp *subRegExp = &reg_exps[sub_index];
	
	NameSet *nameSet = &newOne->nameSet;
	NameSet *subNameSet = &subRegExp->nameSet;
	
	newOne->data.optionalData.subRegExp = sub_index;
	nameSet_addAll(nameSet,subNameSet);
	return index;
}

int newMinMaxOcc(int subRegExp, int min, int max) {
	if(min < 0 || (min > max && max != INFINITE_REPEAT)) {
		synerr( "bad iteration values" );
		return subRegExp;
	}
	if(min == 0) return newMaxOcc(subRegExp,max);
	if(min == max) return newFixOcc(subRegExp,min);
	if(max == INFINITE_REPEAT) return newMinOcc(subRegExp,min);
	
	return createMinMaxOcc(false,max,min,subRegExp);
}

int newMinOcc(int subRegExp, int min) {
	if(min < 0) {
		synerr( "iteration value must be positive" );
		return subRegExp;
	}
	if(min == 0) return newStar(subRegExp);
	if(min == 1) return newPlus(subRegExp);
	
	return createMinMaxOcc(true,min,min,subRegExp);
}

int newMaxOcc(int subRegExp, int max) {
	if(max < 0 && max != INFINITE_REPEAT) {
		synerr( "iteration value must be positive" );
		return subRegExp;
	}
	if(max == 0) return newEpsilon();
	if(max == 1) return newOptional(subRegExp);
	if(max == INFINITE_REPEAT) return newStar(subRegExp);
	
	return createMinMaxOcc(false,max,0,subRegExp);
}

int newFixOcc(int subRegExp, int fixed) {
	if(fixed < 0) {
		synerr( "iteration value must be positive" );
		return subRegExp;
	}
	if(fixed == 0) return newEpsilon();
	if(fixed == 1) return subRegExp;
	
	return createMinMaxOcc(false,fixed,fixed,subRegExp);
}

int newNamed(int sub_index, int name) {
	int index = newRegExp(NAMED);
	RegExp *newOne = &reg_exps[index];
	RegExp *subRegExp = &reg_exps[sub_index];
	
	NameSet *nameSet = &newOne->nameSet;
	NameSet *subNameSet = &subRegExp->nameSet;
	
	if(nameSet_containsName(subNameSet,name)) {
		synerr( "recursive name" );
		return sub_index;
	}
	
	newOne->data.namedData.name = name;
	newOne->data.namedData.subRegExp = sub_index;
	nameSet_addAll(nameSet,subNameSet);
	nameSet_addNameLevel(nameSet,name,0);
	return index;
}

// METHODS

int regExp_createMachine(int index) {
	RegExp *regExp = &reg_exps[index];
	int type = regExp->type;
	int result = -1;
	
	if(type == EPSILON) {
		result = mkstate(SYM_EPSILON);
	}
	else if(type == VALUE) {
		int value = regExp->data.valueData.value;
		int priority = regExp->data.valueData.priority;
		if(rule_uses_actions && priority != 0) result = mkstate_actionList(value,newActionList(priority));
		else result = mkstate(value);
	}
	else if(type == CHAIN) {
		int first = regExp_createMachine(regExp->data.chainData.firstSubRegExp);
		int second = regExp_createMachine(regExp->data.chainData.secondSubRegExp);
		result = link_machines(first,second);
	}
	else if(type == UNION) {
		int first = regExp_createMachine(regExp->data.unionData.firstSubRegExp);
		int second = regExp_createMachine(regExp->data.unionData.secondSubRegExp);
		result = mkor(first,second);
	}
	else if(type == NAMED) {
		result = regExp_createMachine(regExp->data.namedData.subRegExp);
	}
	else if(type == STAR) {
		int subMachine = regExp_createMachine(regExp->data.starData.subRegExp);
		int recursivePriority = rule_uses_actions?regExp->data.starData.recursivePriority:0;
		result = mkclos_recursivePriority(subMachine,recursivePriority);
	}
	else if(type == PLUS) {
		int subMachine = regExp_createMachine(regExp->data.plusData.subRegExp);
		int recursivePriority = rule_uses_actions?regExp->data.plusData.recursivePriority:0;
		result = mkposcl_recursivePriority(subMachine,recursivePriority);
	}
	else if(type == OPTIONAL) {
		result = mkopt(regExp_createMachine(regExp->data.optionalData.subRegExp));
	}
	else if(type == MIN_MAX_OCC) {
		MinMaxOccData* data = &regExp->data.minMaxOccData;
		int infinite = data->infinite;
		int size = data->size;
		int minimum = data->minimum;
		int subRegExp = data->subRegExp;
		
		int i, duplicate, optional;
		int recursivePriorityBase = 0;
		int priority = 0;
		
		if(rule_uses_actions) {
			recursivePriorityBase = data->recursivePriorityBase;
			priority = data->priorityStart + recursivePriorityBase * size;
		}
		
		if(infinite) {
			if(rule_uses_actions) regExp_setPriorities(subRegExp,priority -= recursivePriorityBase);
			duplicate = regExp_createMachine(subRegExp);
			result = mkposcl_recursivePriority(duplicate,recursivePriorityBase);
			minimum--;
			optional = 0;
		}
		else {
			result = mkstate(SYM_EPSILON);
			optional = size-minimum;
		}
		
		for (i=0; i<optional; i++) {
			if(rule_uses_actions) regExp_setPriorities(subRegExp,priority -= recursivePriorityBase);
			duplicate = regExp_createMachine(subRegExp);
			result = mkopt_priority(link_machines(duplicate,result),rule_uses_actions?-recursivePriorityBase*(i+1)+1:0);
		}
		
		for (i=0; i<minimum; i++) {
			if(rule_uses_actions) regExp_setPriorities(subRegExp,priority -= recursivePriorityBase);
			duplicate = regExp_createMachine(subRegExp);
			result = link_machines(duplicate,result);
		}
	}
	else flexfatal(
		_( "unknown regular expression type in \"regExp_createMachine\"" ) );
	
	if(rule_uses_actions) result = machine_addActionTransitions(result,&regExp->actionSet);
	
	return result;
}

void regExp_finish(int regExp) {
	finish_name_options();
	regExp_finishLevels(regExp);
	regExp_finishPriorities(regExp);
}

void regExp_finishLevels(int index) {
	regExp_top_levelManagement(index,UNUSED);
}

void regExp_finishPriorities(int regExp) {
	globalPriorities = regExp_setPriorities(regExp,globalPriorities);
}

///////////////
// PRIVATE
///////////////

// CONSTRUCTORS

int newRegExp(int type) {
	if (reg_exps_size == reg_exps_size_max ) enlarge_reg_exps();
	
	if(reg_exps[reg_exps_size].type == UNUSED) {
		reg_exps[reg_exps_size].nameSet.max_size = INITIAL_NAME_SET_ELEMENT_SIZE;
		reg_exps[reg_exps_size].nameSet.array = allocate_nameLevelCount_array(INITIAL_NAME_SET_ELEMENT_SIZE);
		reg_exps[reg_exps_size].nameSet.size = 0;
		
		reg_exps[reg_exps_size].actionSet.max_size = INITIAL_ACTION_SET_ELEMENT_SIZE;
		reg_exps[reg_exps_size].actionSet.array = allocate_nameLevelRange_array(INITIAL_ACTION_SET_ELEMENT_SIZE);
		reg_exps[reg_exps_size].actionSet.size = 0;
	}
	else {
		reg_exps[reg_exps_size].nameSet.size = 0;
		reg_exps[reg_exps_size].actionSet.size = 0;
	}
	
	reg_exps[reg_exps_size].type = type;
	return reg_exps_size++;
}

int createMinMaxOcc(int infinite, int size, int minimum, int subRegExp) {
	int index = newRegExp(MIN_MAX_OCC);
	regExp_closure_initialLevelManagement(index,subRegExp);
	RegExp *newOne = &reg_exps[index];
	
	newOne->data.minMaxOccData.infinite = infinite;
	newOne->data.minMaxOccData.size = size;
	newOne->data.minMaxOccData.minimum = minimum;
	newOne->data.minMaxOccData.subRegExp = subRegExp;
	
	return index;
}

// METHODS

void regExp_enableActionSet(int index, NameLevel* nameLevelMap, int nameLevelMapSize) {
	RegExp *regExp = &reg_exps[index];
	int type = regExp->type;
	
	if(type == EPSILON) {}
	else if(type == VALUE) {}
	else if(type == CHAIN) {
		regExp_chain_enableActionSet(index,nameLevelMap,nameLevelMapSize);
	}
	else if(type == UNION) {
		regExp_union_enableActionSet(index,nameLevelMap,nameLevelMapSize);
	}
	else if(type == NAMED) {
		regExp_named_enableActionSet(index,nameLevelMap,nameLevelMapSize);
	}
	else if(type == STAR) {
		regExp_closure_enableActionSet(index,regExp->data.starData.subRegExp,nameLevelMap,nameLevelMapSize);
	}
	else if(type == PLUS) {
		regExp_closure_enableActionSet(index,regExp->data.plusData.subRegExp,nameLevelMap,nameLevelMapSize);
	}
	else if(type == OPTIONAL) {
		regExp_enableActionSet(regExp->data.optionalData.subRegExp,nameLevelMap,nameLevelMapSize);
	}
	else if(type == MIN_MAX_OCC) {
		regExp_closure_enableActionSet(index,regExp->data.minMaxOccData.subRegExp,nameLevelMap,nameLevelMapSize);
	}
	else flexfatal(
		_( "unknown regular expression type \"regExp_enableActionSet\"" ) );
}

int regExp_setPriorities(int index, int priority) {
	RegExp *regExp = &reg_exps[index];
	int type = regExp->type;
	
	if(type == EPSILON) {
		return priority;
	}
	else if(type == VALUE) {
		regExp->data.valueData.priority = priority;
		return priority-1;
	}
	else if(type == CHAIN) {
		int nextPriority = regExp_setPriorities(regExp->data.chainData.firstSubRegExp,priority);
		return regExp_setPriorities(regExp->data.chainData.secondSubRegExp,nextPriority);
	}
	else if(type == UNION) {
		int nextPriority = regExp_setPriorities(regExp->data.unionData.firstSubRegExp,priority);
		return regExp_setPriorities(regExp->data.unionData.secondSubRegExp,nextPriority);
	}
	else if(type == NAMED) {
		return regExp_setPriorities(regExp->data.namedData.subRegExp,priority);
	}
	else if(type == STAR) {
		return regExp_closure_setPriorities(index,regExp->data.starData.subRegExp,&regExp->data.starData.recursivePriority,priority);
	}
	else if(type == PLUS) {
		return regExp_closure_setPriorities(index,regExp->data.plusData.subRegExp,&regExp->data.plusData.recursivePriority,priority);
	}
	else if(type == OPTIONAL) {
		return regExp_setPriorities(regExp->data.optionalData.subRegExp,priority);
	}
	else if(type == MIN_MAX_OCC) {
		regExp->data.minMaxOccData.priorityStart = priority;
		regExp_closure_setPriorities(index,regExp->data.minMaxOccData.subRegExp,&regExp->data.minMaxOccData.recursivePriorityBase,priority);
		return priority + regExp->data.minMaxOccData.recursivePriorityBase * regExp->data.minMaxOccData.size - 1;
	}
	
	flexfatal(
		_( "unknown regular expression type \"regExp_setPriorities\"" ) );
	
	return -1;
}

// CLOSURE

void regExp_closure_initialLevelManagement(int closure_index, int subRegExp_index) {
	regExp_top_levelManagement(subRegExp_index,closure_index);
}

void regExp_top_levelManagement(int index, int closure) {
	int i,n;
	RegExp *regExp = &reg_exps[index];
	NameSet *nameSet = &regExp->nameSet;
	ActionSet *actionSet = &regExp->actionSet;
	
	int nameSetSize = nameSet->size;
	
	int usedNamesCount = 0;
	int *usedNames = allocate_integer_array(nameSetSize);
	
	int nameLevelMapSize = 0;
	NameLevel* nameLevelMap = (NameLevel*)allocate_array(nameSetSize,sizeof(NameLevel));
	
	for(i=0; i<nameSetSize; i++) {
		NameLevelCount* nlc = &nameSet->array[i];
		int name = nlc->name, maxLevel, maxCount;
		NameLevelCount *max, *secondMax;
		
		for(n=0; n<usedNamesCount; n++) {
			if(name == usedNames[n]) {
				name = -1;
				break;
			}
		}
		if(name == -1) continue;
		
		usedNames[usedNamesCount++] = name;
		
		max = nameSet_getMaximumLevel(nameSet,name);
		maxLevel = max->level;
		maxCount = max->count;
		
		int levelIncrement = 0;
		if(maxCount > 1 ||
		   (!APPEND(name) && nameSet_getSecondMaximumLevel(nameSet,name) != NIL)) {
			actionSet_addNameLevel(actionSet,name,maxLevel+1,maxLevel+1);
			levelIncrement++;
		}
		
		if(closure != UNUSED) {
			// for closure
			nameSet_addNameLevel(&reg_exps[closure].nameSet,name,maxLevel+levelIncrement+1);
		}
		else {
			// for top level element
			max_level = MAX(max_level,maxLevel+levelIncrement);
			name_dimensions[name] = maxLevel+levelIncrement;
		}
		
		nameLevelMap[nameLevelMapSize].name = name;
		nameLevelMap[nameLevelMapSize++].level = maxLevel;
	}
	
	if(nameLevelMapSize > 0) regExp_enableActionSet(index,nameLevelMap,nameLevelMapSize);
	
	flex_free(usedNames);
	flex_free(nameLevelMap);
}

void regExp_closure_enableActionSet(int closure_index, int subRegExp_index, NameLevel* nameLevelMap, int nameLevelMapSize) {
	int i,n;
	RegExp *closure = &reg_exps[closure_index];
	RegExp *subRegExp = &reg_exps[subRegExp_index];
	
	NameSet* nameSet = &closure->nameSet;
	ActionSet* actionSet = &closure->actionSet;
	
	for(i=0; i<nameLevelMapSize; i++) {
		NameLevel* nl = &nameLevelMap[i];
		int name = nl->name;
		int minLevel, maxLevel;
		
		if(!nameSet_containsName(nameSet,name)) continue;
		
		minLevel = nameSet_getMaximumLevel(nameSet,name)->level;
		maxLevel = nl->level;
		
		if(minLevel <= maxLevel) actionSet_addNameLevel(actionSet,name,minLevel,maxLevel);
	}
}

int regExp_closure_setPriorities(int closure, int subRegExp, int* recursivePriority, int next) {
	int start = next; 
	next = regExp_setPriorities(subRegExp,next);
	*recursivePriority = next - start;
	return next-1;
}

// CHAIN

void regExp_chain_enableActionSet(int index, NameLevel* nameLevelMap, int nameLevelMapSize) {
	int i,n;
	RegExp* chain = &reg_exps[index];
	
	int first_index = chain->data.chainData.firstSubRegExp;
	int second_index = chain->data.chainData.secondSubRegExp;
	
	RegExp* first = &reg_exps[first_index];
	RegExp* second = &reg_exps[second_index];
	
	NameSet* nameSet = &chain->nameSet;
	int nameSetSize = nameSet->size;
	
	NameSet* firstSubNameSet = &first->nameSet;
	int firstSubNameSetSize = firstSubNameSet->size;
	
	NameSet* secondSubNameSet = &second->nameSet;
	int secondSubNameSetSize = secondSubNameSet->size;
	
	int firstNameLevelMapSize = 0;
	NameLevel* firstNameLevelMap = (NameLevel*)allocate_array(firstSubNameSetSize,sizeof(NameLevel));
	
	int secondNameLevelMapSize = 0;
	NameLevel* secondNameLevelMap = (NameLevel*)allocate_array(secondSubNameSetSize,sizeof(NameLevel));
	
	for(i=0; i<nameLevelMapSize; i++) {
		NameLevel* nl = &nameLevelMap[i];
		int name = nl->name;
		int enableLevel = nl->level;
		int maxLevel, maxCount;
		int concatenate, max_equals_enable, firstExist, secondExist;
		NameLevelCount *firstMax,*secondMax;
		
		for(n=0; n<nameSetSize; n++) {
			if(name == nameSet->array[n].name) break;
		}
		if(n == nameSetSize) continue;
		
		firstMax = nameSet_getMaximumLevel(firstSubNameSet,name);
		secondMax = nameSet_getMaximumLevel(secondSubNameSet,name);
		
		firstExist = firstMax != NIL;
		secondExist = secondMax != NIL;
		
		if(firstExist && secondExist) {
			NameLevelCount* max = nameSet_getMaximumLevel(nameSet,name);
			maxLevel = max->level;
			maxCount = max->count;
			
			if(maxLevel-enableLevel > 1) flexfatal(
				_( "cannot enable level in chain" ) );
			
			if(APPEND(name) == NAME_OPTION_YES && maxLevel == enableLevel && firstMax->level != secondMax->level && maxCount == 1) {
				actionSet_addNameLevel(&chain->actionSet,name,enableLevel,enableLevel);
				enableLevel--;
			}
			else if(maxLevel < enableLevel) {
				if(APPEND(name) && maxCount == 1 && nameSet_getSecondMaximumLevel(nameSet,name) != NIL) {
					actionSet_addNameLevel(&chain->actionSet,name,maxLevel,enableLevel);
					enableLevel = maxLevel-1;
				}
				else if(CONCAT(name)) {
					actionSet_addNameLevel(&chain->actionSet,name,maxLevel+1,enableLevel);
					enableLevel = maxLevel;
				}
			}
		}
		
		if(firstExist) {
			firstNameLevelMap[firstNameLevelMapSize].name = name;
			firstNameLevelMap[firstNameLevelMapSize++].level = enableLevel;
		}
		
		if(secondExist) {
			secondNameLevelMap[secondNameLevelMapSize].name = name;
			secondNameLevelMap[secondNameLevelMapSize++].level = enableLevel;
		}
	}
	
	if(firstNameLevelMapSize > 0) regExp_enableActionSet(first_index,firstNameLevelMap,firstNameLevelMapSize);
	if(secondNameLevelMapSize > 0) regExp_enableActionSet(second_index,secondNameLevelMap,secondNameLevelMapSize);
	
	flex_free(firstNameLevelMap);
	flex_free(secondNameLevelMap);
}

// UNION

void regExp_union_enableActionSet(int index, NameLevel* nameLevelMap, int nameLevelMapSize) {
	int i,n;
	RegExp* regExp = &reg_exps[index];
	
	int first_index = regExp->data.unionData.firstSubRegExp;
	int second_index = regExp->data.unionData.secondSubRegExp;
	
	RegExp* first = &reg_exps[first_index];
	RegExp* second = &reg_exps[second_index];
	
	NameSet* nameSet = &regExp->nameSet;
	int nameSetSize = nameSet->size;
	
	NameSet* firstSubNameSet = &first->nameSet;
	int firstSubNameSetSize = firstSubNameSet->size;
	
	NameSet* secondSubNameSet = &second->nameSet;
	int secondSubNameSetSize = secondSubNameSet->size;
	
	int firstNameLevelMapSize = 0;
	NameLevel* firstNameLevelMap = (NameLevel*)allocate_array(firstSubNameSetSize,sizeof(NameLevel));
	
	int secondNameLevelMapSize = 0;
	NameLevel* secondNameLevelMap = (NameLevel*)allocate_array(secondSubNameSetSize,sizeof(NameLevel));
	
	for(i=0; i<nameLevelMapSize; i++) {
		NameLevel* nl = &nameLevelMap[i];
		int name = nl->name;
		
		if(!nameSet_containsName(nameSet,name)) continue;
		
		if(nameSet_containsName(firstSubNameSet,name)) {
			firstNameLevelMap[firstNameLevelMapSize].name = name;
			firstNameLevelMap[firstNameLevelMapSize++].level = nl->level;
		}
		
		if(nameSet_containsName(secondSubNameSet,name)) {
			secondNameLevelMap[secondNameLevelMapSize].name = name;
			secondNameLevelMap[secondNameLevelMapSize++].level = nl->level;
		}
	}
	
	if(firstNameLevelMapSize > 0) regExp_enableActionSet(first_index,firstNameLevelMap,firstNameLevelMapSize);
	if(secondNameLevelMapSize > 0) regExp_enableActionSet(second_index,secondNameLevelMap,secondNameLevelMapSize);
	
	flex_free(firstNameLevelMap);
	flex_free(secondNameLevelMap);
}

// NAMED

void regExp_named_enableActionSet(int index, NameLevel* nameLevelMap, int nameLevelMapSize) {
	int i,n;
	
	RegExp* named = &reg_exps[index];
	NameSet* nameSet = &named->nameSet;
	ActionSet* actionSet = &named->actionSet;
	
	int sub_index = named->data.namedData.subRegExp;
	RegExp* subRegExp = &reg_exps[sub_index];
	
	NameSet* subNameSet = &subRegExp->nameSet;
	int subNameSetSize = subNameSet->size;
	
	int subNameLevelMapSize = 0;
	NameLevel* subNameLevelMap = (NameLevel*)allocate_array(subNameSetSize,sizeof(NameLevel));
	
	for(i=0; i<nameLevelMapSize; i++) {
		NameLevel* nl = &nameLevelMap[i];
		int name = nl->name;
		
		if(!nameSet_containsName(nameSet,name)) continue;
		
		if(name == named->data.namedData.name) {
			int maxLevel = nl->level;
			actionSet_addNameLevel(actionSet,name,0,maxLevel);
		}
		else {
			subNameLevelMap[subNameLevelMapSize].name = name;
			subNameLevelMap[subNameLevelMapSize++].level = nl->level;
		}
	}
	
	if(subNameLevelMapSize > 0) regExp_enableActionSet(sub_index,subNameLevelMap,subNameLevelMapSize);
	
	flex_free(subNameLevelMap);
}

// Actions

int machine_addActionTransitions(int machine, ActionSet *actionSet) {
	NameLevelRange *ranges = actionSet->array;
	int i,n,rangesSize = actionSet->size;
	int enter,exit;
	int exitNecessary = false;
	
	if(rangesSize == 0) return machine;
	
	enter = mkstate(SYM_EPSILON);
	machine = link_machines_actionList(enter,newActionList(0),machine);
	
	for(i=0; i<rangesSize; i++) {
		int name = ranges[i].name;
		int min = ranges[i].lowestLevel;
		int max = ranges[i].highestLevel;
		
		for(n=max; n>=min; n--) {
			addActionToLastCreatedActionList(name,n,true);
		}
		
		if(min == 0) exitNecessary = true;
	}
	
	if(exitNecessary) {
		int finalState = finalst[machine];
		
		if(!SUPER_FREE_EPSILON(finalState)) {
			finalState = mkstate(SYM_EPSILON);
			machine = link_machines(machine,finalState);
		}
		
		finalState = mkstate(SYM_EPSILON);
		machine = link_machines_actionList(machine,newActionList(0),finalState);
		
		for(i=0; i<rangesSize; i++) {
			int name = ranges[i].name;
			int min = ranges[i].lowestLevel;
			//int max = ranges[i].highestLevel;
			
			if(min == 0) addActionToLastCreatedActionList(name,0,false);
		
			/*for(n=min; n<=max; n++) {
				addActionToLastCreatedActionList(name,n,false);
			}*/
		}
	}
	
	return machine;
}

void finish_name_options() {
	int i;
	
	for(i=0; i<rule_names_size; i++) {
		int nameIndex = rule_names[i];
		char *name_string = names[nameIndex];
		
		finish_name_option(&name_append[nameIndex],"append",name_string,append);
		finish_name_option(&name_concat[nameIndex],"concat",name_string,concat);
	}
	
	name_options_initialized = false;
}

void finish_name_option(int* option_value, char* option_name, char* name, int default_value) {
	
	switch(*option_value) {
	
	case NAME_OPTION_REDUNDANT_NO:
		fprintf(stderr,"warning: redundant name option no%s for name \"%s\" in rule at line %i",
			option_name,name,linenum);
		*option_value = NAME_OPTION_NO;
		break;
		
	case NAME_OPTION_REDUNDANT_YES:
		fprintf(stderr,"warning: redundant name option %s for name \"%s\" in rule at line %i",
			option_name,name,linenum);
		*option_value = NAME_OPTION_YES;
		break;
		
	case NAME_OPTION_CONFLICTING:
		fprintf(stderr,"warning: conflicting name options %s and no%s for name \"%s\"\n\tin rule at line %i -> using default value %s%s",
			option_name,option_name,name,linenum,default_value?"":"no",option_name);
		*option_value = NAME_OPTION_DEFAULT;
		break;
		
	case NAME_OPTION_REDUNDANT_AND_CONFLICTING:
		fprintf(stderr,"warning: redundant and conflicting name options %s and no%s for name \"%s\"\n\tin rule at line %i -> using default value %s%s",
			option_name,option_name,name,linenum,default_value?"":"no",option_name);
		*option_value = NAME_OPTION_DEFAULT;
		break;
	}
	
	if(*option_value == NAME_OPTION_DEFAULT)
		*option_value = default_value?NAME_OPTION_YES:NAME_OPTION_NO;
}


/**********************************************************
 * Methods for NameSet
 **********************************************************/

void nameSet_addNameLevelCount_unchecked PROTO((NameSet*,int,int,int));

///////////////
// PUBLIC
///////////////

void nameSet_addNameLevel(NameSet *nameSet, int name, int level) {
	nameSet_addNameLevelCount(nameSet,name,level,1);
}

void nameSet_addNameLevelCount(NameSet *nameSet, int name, int level, int count) {
	int size = nameSet->size;
	int i;
	
	for(i=0; i<size; i++) {
		NameLevelCount *nlc = &nameSet->array[i];
		if(nlc->name == name &&
		   nlc->level == level)
		{
		   nlc->count = nlc->count + count;
		   return;
		}
	}
	
	nameSet_addNameLevelCount_unchecked(nameSet,name,level,count);
}

void nameSet_setToMaximumCount(NameSet *nameSet, int name, int level, int count) {
	int size = nameSet->size;
	int i;
	
	for(i=0; i<size; i++) {
		NameLevelCount *nlc = &nameSet->array[i];
		if(nlc->name == name &&
		   nlc->level == level)
		{
		   nlc->count = MAX(nlc->count,count);
		   return;
		}
	}
	
	nameSet_addNameLevelCount_unchecked(nameSet,name,level,count);
}

void nameSet_addAll(NameSet *nameSet, NameSet *otherNameSet) {
	int size = otherNameSet->size;
	int i;
	
	for(i=0; i<size; i++) {
		NameLevelCount* nlc = &otherNameSet->array[i];
		nameSet_addNameLevelCount(nameSet,nlc->name,nlc->level,nlc->count);
	}
}

void nameSet_addAll_ofName(NameSet *nameSet, NameSet *otherNameSet, int name) {
	int size = otherNameSet->size;
	int i;
	
	for(i=0; i<size; i++) {
		NameLevelCount* nlc = &otherNameSet->array[i];
		int other_name = nlc->name;
		if(other_name == name) nameSet_addNameLevelCount(nameSet,name,nlc->level,nlc->count);
	}
}

void nameSet_setAllToMaximumCount(NameSet *nameSet, NameSet *otherNameSet) {
	int size = nameSet->size;
	int i;
	
	for(i=0; i<size; i++) {
		NameLevelCount* nlc = &otherNameSet->array[i];
		nameSet_setToMaximumCount(nameSet,nlc->name,nlc->level,nlc->count);
	}
}

NameLevelCount* nameSet_getMaximumLevel(NameSet *nameSet, int name) {
	int size = nameSet->size;
	int i, maxLevel = -1;
	NameLevelCount * result = NIL;
	
	for(i=0; i<size; i++) {
		NameLevelCount* nlc = &nameSet->array[i];
		if(nlc->name == name && nlc->level > maxLevel) {
			result = nlc;
			maxLevel = nlc->level;
		}
	}
	
	return result;
}

NameLevelCount* nameSet_getSecondMaximumLevel(NameSet *nameSet, int name) {
	int size = nameSet->size;
	int i, secondMaxLevel = -1, maxLevel = -1;
	NameLevelCount *max = NIL, *secondMax = NIL;
	
	for(i=0; i<size; i++) {
		NameLevelCount* nlc = &nameSet->array[i];
		if(nlc->name == name) {
			int level = nlc->level;
			if(level > maxLevel) {
				secondMaxLevel = maxLevel;
				secondMax = max;
				
				maxLevel = level;
				max = nlc;
			}
			else if(level > secondMaxLevel) {
				secondMaxLevel = level;
				secondMax = nlc;
			}
		}
	}
	
	return secondMax;
}

int nameSet_getCount(NameSet *nameSet, int name, int level) {
	int size = nameSet->size;
	int i;
	
	for(i=0; i<size; i++) {
		NameLevelCount* nlc = &nameSet->array[i];
		if(nlc->name == name && nlc->level == level) return nlc->count;
	}
	
	return 0;
}

int nameSet_containsName(NameSet *nameSet, int name) {
	int size = nameSet->size;
	int i;
	
	for(i=0; i<size; i++) if(nameSet->array[i].name == name) return true;
	return false;
}

///////////////
// PRIVATE
///////////////

void nameSet_addNameLevelCount_unchecked(NameSet *nameSet, int name, int level, int count) {
	if (nameSet->size == nameSet->max_size ) {
		++num_reallocs;
		nameSet->max_size += MAX_NAME_SET_ELEMENT_SIZE_INCREMENT;
		nameSet->array = reallocate_nameLevelCount_array( nameSet->array, nameSet->max_size);
	}
	
	NameLevelCount *nlc = &nameSet->array[nameSet->size];
	nlc->name = name;
	nlc->level = level;
	nlc->count = count;
	nameSet->size = nameSet->size + 1;
}



/**********************************************************
 * Methods for ActionSet
 **********************************************************/

///////////////
// PUBLIC
///////////////

void actionSet_addNameLevel(ActionSet *actionSet, int name, int lowestLevel, int highestLevel) {
	if(actionSet->size == actionSet->max_size) {
		++num_reallocs;
		actionSet->max_size += MAX_NAME_SET_ELEMENT_SIZE_INCREMENT;
		actionSet->array = reallocate_nameLevelRange_array(actionSet->array,actionSet->max_size);
	}
	
	NameLevelRange *nlr = &actionSet->array[actionSet->size];
	nlr->name = name;
	nlr->lowestLevel = lowestLevel;
	nlr->highestLevel = highestLevel;
	actionSet->size = actionSet->size + 1;
}



/**********************************************************
 * Methods for ActionList
 **********************************************************/

///////////////
// PUBLIC
///////////////

int newActionList(int priority) {
	if(action_lists_size == action_lists_size_max) {
		++num_reallocs;
		action_lists_size_max += MAX_ACTION_LISTS_SIZE_INCREMENT;
		action_lists = reallocate_actionList_array(action_lists,action_lists_size_max);
	}
	
	ActionList *newActionList = &action_lists[action_lists_size];
	newActionList->length = 0;
	newActionList->priority = priority;
	newActionList->array = UNUSED;
	
	return action_lists_size++;
}

int newActionListFromOne(int one) {
	int i, start, end;
	int result;
	
	if(one == UNUSED) return UNUSED; //return newActionList(0);
	
	start = action_lists[one].array;
	end = start + action_lists[one].length;
	
	result = newActionList(action_lists[one].priority);
	
	for(i=start; i<end; i++) {
		Action* action = &actions[i];
		addActionToLastCreatedActionList(action->name,action->level,action->sign);
	}
	
	return result;
}

int newActionListFromTwo(int first, int second) {
	int i, firstStart, firstEnd, secondStart, secondEnd;
	int result;
	
	if(first == UNUSED) return newActionListFromOne(second);
	if(second == UNUSED) return newActionListFromOne(first);
	
	firstStart = action_lists[first].array;
	firstEnd = firstStart + action_lists[first].length;
	
	secondStart = action_lists[second].array;
	secondEnd = secondStart + action_lists[second].length;
	
	result = newActionList(action_lists[first].priority + action_lists[second].priority);
	
	for(i=firstStart; i<firstEnd; i++) {
		Action* action = &actions[i];
		addActionToLastCreatedActionList(action->name,action->level,action->sign);
	}
	
	for(i=secondStart; i<secondEnd; i++) {
		Action* action = &actions[i];
		addActionToLastCreatedActionList(action->name,action->level,action->sign);
	}
	
	return result;
}

int newActionListFromActionListReferenceList(int list) {
	int i,n, size, priority = 0;
	int result;
	
	ActionListReferenceList* origin = &actionListReferenceLists[list];
	size = origin->size;
	
	for(i=0; i<size; i++) priority += action_lists[origin->elements[i]].priority;
	
	result = newActionList(priority);
	
	for(i=0; i<size; i++) {
		ActionList* actionList = &action_lists[origin->elements[i]];
		int start = actionList->array;
		int end = start + actionList->length;
		
		for(n=start; n<end; n++) {
			Action* action = &actions[n];
			addActionToLastCreatedActionList(action->name,action->level,action->sign);
		}
	}
	
	return result;
}

void addActionToLastCreatedActionList(int name, int level, int sign) {
	ActionList *lastActionList = &action_lists[action_lists_size-1];
	
	if(actions_size == actions_size_max) {
		++num_reallocs;
		actions_size_max += MAX_ACTIONS_SIZE_INCREMENT;
		actions = reallocate_action_array(actions,actions_size_max);
	}
	
	Action *newAction = &actions[actions_size++];
	newAction->name = name;
	newAction->level = level;
	newAction->sign = sign;
	
	if(lastActionList->array == UNUSED) lastActionList->array = actions_size-1;
	
	lastActionList->length = lastActionList->length + 1;
}

void action_print(Action* action, FILE* file) {
	char sign[2];
	sign[1] = 0;
	
	if(action->sign) sign[0] = '+';
	else sign[0] = '-';
	
	if(action->level == 0) fprintf(file,"%s%s",sign,names[action->name]);
	else fprintf(file,"%s%s[%i]",sign,names[action->name],action->level);
}

void actionList_print(int actionList, FILE* file) {
	int i, start, end, priority;
	
	if(actionList == UNUSED) {
		fprintf(file,"UNUSED");
		return;
	}
	
	start = action_lists[actionList].array;
	end = start + action_lists[actionList].length;
	priority = action_lists[actionList].priority;
	
	fprintf(file,"[");
	
	for(i=start;i<end;i++) {
		if(i!=start) fprintf(file,",");
		action_print(&actions[i],file);
	}
	
	fprintf(file,"]");
	if(priority != 0) fprintf(file,"!!%i",priority);
}
